/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2004-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl>
 *
 */



#include "verify_x509.h"
#include "_verify_x509.h"
#include "log.h"



int verify_X509_init (internal_verify_x509_data_t ** verify_x509_data)
{
    OpenSSL_add_all_algorithms();
    ERR_load_crypto_strings();

    *verify_x509_data = calloc (1, sizeof (internal_verify_x509_data_t));
    
    return *verify_x509_data ? 0 : 1;
}

int verify_X509_setParameter (internal_verify_x509_data_t ** verify_x509_data,  int verify_x509_option, ...)
{
    internal_verify_x509_data_t * internal_verify_x509_data = *verify_x509_data;
    struct stat       my_stat;
    unsigned long     result = 0;
    va_list           ap;


    va_start (ap, verify_x509_option);
    switch (verify_x509_option)
    {
        case VERIFY_X509_CA_PATH :
        {
            char * data_capath = NULL;
        
            if (internal_verify_x509_data -> capath)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }

            /* cast arb_data */
            data_capath = va_arg (ap, char *);

            /* Check of not NULL */
            if (!data_capath)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_DATA_EMPTY; 
            }
             
            /* Check if we can access the data */
            if ( (result = stat (data_capath, &my_stat ) ) != 0 )
            {
                Error ("Error: unable to stat() CA directory:", "%s", data_capath);
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ACCESS_FAILURE;
            }

            internal_verify_x509_data -> capath                 = strdup (data_capath);
            break;
        }

        case VERIFY_X509_CERTIFICATE_FILEPATH :
        {
            char * data_filepath = NULL;
        
            if (internal_verify_x509_data -> certificate_filepath)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }

            /* cast arb_data */
            data_filepath = va_arg (ap, char *);

            /* Check of not NULL */
            if (!data_filepath)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_DATA_EMPTY; 
            }
                
            /* Check if we can access the data */
            if ( (result = stat (data_filepath, &my_stat ) ) != 0 )
            {
                Error ("Error: unable to stat() Certificate File:", "%s", data_filepath);
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ACCESS_FAILURE;
            }

            internal_verify_x509_data -> certificate_filepath   = strdup (data_filepath);
            break;
        }

        case VERIFY_X509_CERTIFICATE_F_HANDLE :
            if (internal_verify_x509_data -> certificate_f_handle)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }
            internal_verify_x509_data -> certificate_f_handle   = va_arg (ap, FILE *);
            break;
        case VERIFY_X509_CERTIFICATE_PEM :
            if (internal_verify_x509_data -> certificate_pem_str)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }
            internal_verify_x509_data -> certificate_pem_str    = strdup (va_arg (ap, char *));
            break;
        case VERIFY_X509_PRIVATEKEY_FILE :
            if (internal_verify_x509_data -> private_key_filepath)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }
            internal_verify_x509_data -> private_key_filepath   = strdup (va_arg (ap, char *));

            if ( (result = stat (internal_verify_x509_data -> private_key_filepath, &my_stat ) ) != 0 )
            {
                Error ("Error: unable to stat() Private Key File:", "%s", internal_verify_x509_data -> private_key_filepath);
                free (internal_verify_x509_data -> private_key_filepath);
                internal_verify_x509_data -> private_key_filepath = NULL;
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ACCESS_FAILURE;
            }

            break;
        case VERIFY_X509_PRIVATEKEY_PEM :
            if (internal_verify_x509_data -> private_key_pem)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }
            internal_verify_x509_data -> private_key_pem        = strdup (va_arg (ap, char *));
            break;
        case VERIFY_X509_CRL_PATH :
            if (internal_verify_x509_data -> crl_path)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }
            internal_verify_x509_data -> crl_path               = strdup (va_arg (ap, char *));
            break;
        case VERIFY_X509_OSPF_RESPONDER_URI : 
            if (internal_verify_x509_data -> ospf_responder_uri)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }
            internal_verify_x509_data -> ospf_responder_uri     = strdup (va_arg (ap, char *));
            break;


        case VERIFY_X509_OPTIONS_NO_CRL_CHECK :
            internal_verify_x509_data -> no_crl_check           = va_arg (ap, unsigned int);
            break;
        case VERIFY_X509_OPTIONS_ALLOW_LIMITED_PROXY :
            internal_verify_x509_data -> allow_limited_proxy    = va_arg (ap, unsigned int);
            break;
        case VERIFY_X509_OPTIONS_DISALLOW_LIMITED_PROXY :
            internal_verify_x509_data -> disallow_limited_proxy = va_arg (ap, unsigned int);
            break;
        case VERIFY_X509_OPTIONS_MUST_HAVE_PRIV_KEY:
            internal_verify_x509_data -> must_have_priv_key     = va_arg (ap, unsigned int);
            break;


        case VERIFY_X509_STACK_OF_X509 :
            if (internal_verify_x509_data -> stack_of_x509)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }
            internal_verify_x509_data -> stack_of_x509          =  va_arg (ap, STACK_OF (X509) *);
            break;
        case VERIFY_X509_EVP_PKEY      :
            if (internal_verify_x509_data -> evp_pkey)
            {
                va_end(ap);
                return ERR_VERIFY_X509_PARAMS_ALREADY_SET;
            }
            internal_verify_x509_data -> evp_pkey               =  va_arg (ap, EVP_PKEY *);
            break;

        default :
            Error ("Unsupported datatype option specified", "%s\n", "the datatype and data specified is not supported and will not be used in the process");
            va_end(ap);
            return ERR_VERIFY_X509_PARAMS_UNSUPPORTED_DATATYPE;
            break;
    }
    va_end(ap);

    return 0;
}


/* Process internal verification data to get to the fundanmental types e.g. STACK_OF (X509) * and EVP_PKEY * */
unsigned long process_internal_verify_data (internal_verify_x509_data_t ** verify_x509_data)
{
    internal_verify_x509_data_t * internal_verify_x509_data = *verify_x509_data;
    unsigned long     result = 0;


    /* If the private key is already set, don't look for it */
    if (!(internal_verify_x509_data -> evp_pkey))
    {
        /* To EVP Private Key from Source Private Key PEM String */
        if (internal_verify_x509_data -> private_key_pem)
        {
            if (X509_V_OK != (result = verify_x509_readPrivateKeyFromPEM  ( internal_verify_x509_data -> private_key_pem,
                                                                           &(internal_verify_x509_data -> evp_pkey))))
            {
                Error ("Failed to read the private key in file:", "%s\n", internal_verify_x509_data -> certificate_filepath);
                return result;
            }
        }

        /* To EVP Private Key from Source Certificate PEM String */
        if (internal_verify_x509_data -> certificate_pem_str)
        {
            if (X509_V_OK != (result = verify_x509_readPrivateKeyFromPEM  ( internal_verify_x509_data -> certificate_pem_str,
                                                                           &(internal_verify_x509_data -> evp_pkey))))
            {
                Error ("Failed to read the private key in file:", "%s\n", internal_verify_x509_data -> certificate_filepath);
                return result;
            }
        }

        /* To EVP Private Key from Source Private Key File at path */
        else if (internal_verify_x509_data -> private_key_filepath)
        {
            if (X509_V_OK != (result = verify_x509_readPrivateKeyFromFile ( internal_verify_x509_data -> private_key_filepath,
                                                                           &(internal_verify_x509_data -> evp_pkey))))
            {
                Error ("Failed to read the private key in file:", "%s\n", internal_verify_x509_data -> certificate_filepath);
                return result;
            }
        }

        /* To EVP Private Key from Source Certificate File at path */
        else if (internal_verify_x509_data -> certificate_filepath)
        {
            if (X509_V_OK != (result = verify_x509_readPrivateKeyFromFile ( internal_verify_x509_data -> certificate_filepath,
                                                                           &(internal_verify_x509_data -> evp_pkey))))
            {
                Error ("Failed to read the private key in file:", "%s\n", internal_verify_x509_data -> certificate_filepath);
                return result;
            }
        }
    }


    /* If the certificate stack is already provided, don't go look for it in the certificate file */
    if (!(internal_verify_x509_data -> stack_of_x509))
    {
        if (X509_V_OK != (verify_x509_readPublicCertChain ( internal_verify_x509_data -> certificate_filepath,
                                                           &(internal_verify_x509_data -> stack_of_x509))))
        {
            Error ("Failed to read the certificate stack in file:", "%s\n", internal_verify_x509_data -> certificate_filepath);
            return result;
        }
    }

    return X509_V_OK;
}



/******************************************************************************
Function: 
    verify_X509_verify
    
Description:
    This is the verification function that will first process the provided 
    setParameters into usable datatypes for the actual verification. Then the 
    verification of the certificate chain (includes GT2 and GT4/RFC proxies) 
    will start. First the chain will be checked and then (if there is a private 
    key) the private key must match the certificate chain.
    
Parameters:
    internal_verify_x509_data_t ** verify_x509_data

Return
    X509_V_OK  :  all is perfect
    !X509_V_OK :  error code
******************************************************************************/
int verify_X509_verify (internal_verify_x509_data_t ** verify_x509_data)
{
    internal_verify_x509_data_t * internal_verify_x509_data = *verify_x509_data;
    unsigned long     result = 0;

    /*
     * Crunch on the supplied information to get to the principle pieces of information required 
     * for the verification based on the STACK_OF (X509) * and the EVP_PKEY *
     */
    if (X509_V_OK != (result = process_internal_verify_data(&internal_verify_x509_data)))
    {
        return result;
    }


    if (result == X509_V_OK)
    {
        /* first, verify the certificate chain */
        result = grid_verifyCert (internal_verify_x509_data -> capath, 
                                  internal_verify_x509_data -> stack_of_x509);
        if (result == X509_V_OK)
        {
            /* When set to have the private key, then this means that the certificate chain 
             * MUST be accompanied by a private key (and in other test this one must fit 
             * the chain)
             */
            if ((internal_verify_x509_data -> must_have_priv_key == VERIFY_ENABLE) &&
                (internal_verify_x509_data -> evp_pkey == NULL))
            {
                Error ("No private key provided", 
                       "the configuration (by default or by explict statement) demands its presence\n"); 
                return VERIFY_X509_OPTIONS_MUST_HAVE_PRIV_KEY;
            }    
            

            if (internal_verify_x509_data -> evp_pkey)
            {
                /* still OK? then match the proxy public and private keys */
                result = grid_verifyPrivateKey (internal_verify_x509_data -> stack_of_x509, internal_verify_x509_data -> evp_pkey);
                if (result == X509_V_OK )
                {
                    /* aaah, nirwana */
                    Log (L_INFO, "Verification of chain with private key is OK\n");
                }
                else
                {
                    Error( "Verifying private key", "%s\n",
                            ERR_reason_error_string( result ) );
                }
            }
            else
            {
                Log (L_INFO, "Verification of chain without private key is OK\n");
            }
        }
        else
        {
            Error( "Verifying certificate chain", "%s\n",
                   X509_verify_cert_error_string( result ) );
        }
    }

    return result;
}


int verify_X509_term (internal_verify_x509_data_t ** verify_x509_data)
{
    internal_verify_x509_data_t * internal_verify_x509_data = *verify_x509_data;

    if (internal_verify_x509_data -> capath)
        free (internal_verify_x509_data -> capath);
        
    if (internal_verify_x509_data -> certificate_filepath)
        free (internal_verify_x509_data -> certificate_filepath);
        
    if (internal_verify_x509_data -> certificate_f_handle)
        internal_verify_x509_data -> certificate_f_handle = NULL;
        
    if (internal_verify_x509_data -> certificate_pem_str)
        free (internal_verify_x509_data -> certificate_pem_str);

    if (internal_verify_x509_data -> private_key_filepath)
        free (internal_verify_x509_data -> private_key_filepath);

    if (internal_verify_x509_data -> private_key_pem)
        free (internal_verify_x509_data -> private_key_pem);

    if (internal_verify_x509_data -> crl_path)
        free (internal_verify_x509_data -> crl_path);
        
    if (internal_verify_x509_data -> ospf_responder_uri)
        free (internal_verify_x509_data -> ospf_responder_uri);

    internal_verify_x509_data -> no_crl_check            = 0;
    internal_verify_x509_data -> allow_limited_proxy     = 0;
    internal_verify_x509_data -> disallow_limited_proxy  = 0;
    internal_verify_x509_data -> must_have_priv_key      = 0;

    free(internal_verify_x509_data);
    internal_verify_x509_data = NULL;
    *verify_x509_data         = NULL;
    
    return 0;
}
